from __future__ import annotations

import sqlite3
from pathlib import Path
from inspect import cleandoc
from contextlib import contextmanager

from pydantic import BaseModel


class Session:
    _TABLE_NAME = 'session'

    id: int
    name: str
    model: str
    temperature: float

    @classmethod
    @property
    def ddl(cls):
        return cleandoc(f"""
        CREATE TABLE IF NOT EXISTS {cls._TABLE_NAME} (
            id INTEGER PRIMARY KEY,
            name TEXT NOT NULL UNIQUE,
            model TEXT NOT NULL,
            temperature REAL NOT NULL
        );
        """)


class Message:
    _TABLE_NAME = 'messages'

    id: int
    role: str
    content: str
    session_id: int

    @classmethod
    @property
    def ddl(cls):
        return cleandoc(f"""
        CREATE TABLE IF NOT EXISTS {cls._TABLE_NAME} (
            id INTEGER PRIMARY KEY,
            role TEXT NOT NULL,
            content TEXT NOT NULL,
            session_id INTEGER,
            FOREIGN KEY (session_id) REFERENCES sessions(id)
        );
        """)


class Image(BaseModel):
    _TABLE_NAME = 'images'
    name: str
    data: bytes

    @classmethod
    @property
    def ddl(cls):
        return cleandoc(f"""
        CREATE TABLE  IF NOT EXISTS {cls._TABLE_NAME.default} (
            id INTEGER PRIMARY KEY,
            name TEXT NOT NULL,
            blob BLOB NOT NULL
        );
        """)


TABLES = [Image, Session, Message]


def application_data_dir(name: str) -> Path:
    return Path.home() / '.local' / 'share' / name


def application_db(name: str) -> Path:
    return application_data_dir(name) / f"{name}.sqlite"


@contextmanager
def store(name):
    data_dir = application_data_dir(name)
    data_dir.mkdir(exist_ok=True, parents=True)
    db = application_db(name)
    with sqlite3.connect(db) as con:
        yield con


@contextmanager
def db(name, tables):
    with store(name) as con:
        for table in tables:
            con.execute(table.ddl)
        yield con



import asyncio
import sys
import logging

from functools import wraps
from aergia._cli._parser import make_parser
from aergia._cli._output import stderr, stdout
from aergia._cli._command import ExitCode
from aergia._logging import logger
from rich.logging import RichHandler


def _protect(func, *args, **kwargs):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            exit_code = func(*args, **kwargs)
        except Exception as ex:
            err_msg = f"Error occurred, details: {ex}"
            logger.error(err_msg)
            stdout.print(err_msg, style='red')
            exit_code = ExitCode.Failure
        except asyncio.CancelledError as ex:
            err_msg = f"Canceled by user, details: {ex}"
            logger.error(err_msg)
            stdout.print(err_msg, style='red')
            exit_code = ExitCode.Failure
        return exit_code

    return wrapper


def main(argv=None):
    parser = make_parser()
    args = parser.parse_args(argv)

    if args.log_level:
        level = {
            'debug': logging.DEBUG,
            'info': logging.INFO,
            'warn': logging.WARN,
            'error': logging.ERROR,
            'critical': logging.CRITICAL
        }
        logging.basicConfig(
            level=level[args.log_level],
            format='%(asctime)s - %(name)s - %(levelname)s - %(message)s',
            datefmt='%Y-%m-%d %H:%M:%S',
            handlers=[RichHandler(console=stderr.stderr, rich_tracebacks=True)]
        )

    if not hasattr(args, "func"):
        parser.error("Subcommand required!")

    app = args.func if args.debug else _protect(args.func)
    exit_code = app(args, stdout, stderr)
    sys.exit(exit_code)


if __name__ == "__main__":
    main()



import argparse
from aergia._cli._command import default, chat, image, models
from rich_argparse import ArgumentDefaultsRichHelpFormatter
from contextlib import contextmanager


@contextmanager
def add_parser(subparsers, name, help):
    subparser = subparsers.add_parser(
        name,
        help=help,
        formatter_class=ArgumentDefaultsRichHelpFormatter
    )
    yield subparser


def add_chat_subcommand(subparsers):
    subcommand = subparsers.add_parser(
        'chat',
        help='chat with the ai system',
        formatter_class=ArgumentDefaultsRichHelpFormatter
    )
    subcommand.add_argument(
        'text', type=str, nargs=argparse.REMAINDER, help='Text input'
    )
    subcommand.add_argument(
        '-r', '--role', type=str, help='Choose a role'
    )
    subcommand.add_argument(
        '-s', '--session', nargs='?', const=True, help='Create or reuse a session'
    )
    subcommand.add_argument(
        '-c', '--context', type=argparse.FileType('r'), help='File directory or stdin'
    )
    subcommand.add_argument(
        '-m', '--model',
        default='gpt-4o',
        choices=['gpt-4o', 'gpt-4', 'gpt-4-turbo', 'gpt-3.5-turbo'],
        help='Select a model'
    )
    subcommand.add_argument(
        '--stream', dest='stream', action='store_true', default=True, help='Enable streaming'
    )
    subcommand.set_defaults(func=chat)
    return subcommand


def add_tui_subcommand(subparsers):
    subcommand = subparsers.add_parser(
        'tui',
        help='start textul based tui client',
        formatter_class=ArgumentDefaultsRichHelpFormatter
    )
    subcommand.set_defaults(func=default)
    return subcommand


def add_repl_subcommand(subparsers):
    subcommand = subparsers.add_parser(
        'repl',
        help='start repl client',
        formatter_class=ArgumentDefaultsRichHelpFormatter
    )
    subcommand.set_defaults(func=default)
    return subcommand


def add_code_subcommand(subparsers):
    subcommand = subparsers.add_parser(
        'code',
        help='Generate code',
        formatter_class=ArgumentDefaultsRichHelpFormatter
    )
    subcommand.add_argument(
        'text', type=str, nargs=argparse.REMAINDER, help='Code prompt'
    )
    subcommand.add_argument(
        '-r', '--role', type=str, help='Role to be used in addition'
    )
    subcommand.set_defaults(func=default)
    return subcommand


def add_image_subcommand(subparsers):
    subcommand = subparsers.add_parser(
        'image',
        help='generate images',
        formatter_class=ArgumentDefaultsRichHelpFormatter
    )
    subcommand.add_argument(
        'filename', type=str, help='Name of the imag for storage input'
    )
    subcommand.add_argument(
        '-m', '--model',
        default='dall-e-2',
        choices=['dall-e-2', 'dall-e-3'],
        help='Select a model'
    )
    subcommand.add_argument(
        'text', type=str, nargs=argparse.REMAINDER, help='Text input'
    )
    subcommand.set_defaults(func=image)
    return subcommand


def add_execute_subcommand(subparsers):
    subcommand = subparsers.add_parser(
        'execute',
        help='Generate and execute a command',
        formatter_class=ArgumentDefaultsRichHelpFormatter
    )
    subcommand.add_argument(
        'text', type=str, nargs=argparse.REMAINDER, help='Command prompt'
    )
    subcommand.add_argument(
        '-s', '--shell', type=str, help='Specify the shell to be used for execution'
    )
    subcommand.add_argument(
        '-r', '--role', type=str, help='Role to be taken into account'
    )
    subcommand.set_defaults(func=default)
    return subcommand


def add_session_subcommand(subparsers):
    subcommand = subparsers.add_parser(
        'session',
        help='Manage sessions',
        formatter_class=ArgumentDefaultsRichHelpFormatter
    )
    subcommand.set_defaults(func=default)

    sub_subparsers = subcommand.add_subparsers(
        dest='session_subcommand', help='Sessions commands'
    )

    search_command = sub_subparsers.add_parser(
        'search',
        help='Search for text in sessions',
        formatter_class=ArgumentDefaultsRichHelpFormatter
    )
    search_command.add_argument('text', type=str, help='Text to search')

    list_command = sub_subparsers.add_parser(
        'list',
        help='List all sessions',
        formatter_class=ArgumentDefaultsRichHelpFormatter
    )

    show_command = sub_subparsers.add_parser(
        'show',
        help='Show contents of a session',
        formatter_class=ArgumentDefaultsRichHelpFormatter
    )

    alias_command = sub_subparsers.add_parser(
        'alias',
        help='Alias for session command',
        formatter_class=ArgumentDefaultsRichHelpFormatter
    )
    return subcommand


def add_role_subcommand(subparsers):
    subcommand = subparsers.add_parser(
        'role',
        help='Manage roles',
        formatter_class=ArgumentDefaultsRichHelpFormatter
    )
    subcommand.set_defaults(func=default)

    sub_subparsers = subcommand.add_subparsers(
        dest='role_subcommand', help='Roles commands'
    )

    list_command = sub_subparsers.add_parser(
        'list', help='List all roles'
    )

    show_command = sub_subparsers.add_parser(
        'show', help='Show a role'
    )
    show_command.add_argument(
        'name', type=str, help='Role name'
    )

    create_command = sub_subparsers.add_parser(
        'create', help='Create a role'
    )
    create_command.add_argument(
        'name', type=str, help='Role name'
    )
    create_command.add_argument(
        'text', type=str, nargs='?', help='Role description'
    )

    update_command = sub_subparsers.add_parser(
        'update', help='Update a role'
    )
    update_command.add_argument(
        'name', type=str, help='Role name'
    )
    update_command.add_argument(
        'text', type=str, nargs='?', help='New role description'
    )

    delete_command = sub_subparsers.add_parser(
        'delete', help='Delete a role'
    )
    delete_command.add_argument(
        'name', type=str, help='Role name'
    )

    alias_command = sub_subparsers.add_parser(
        'alias', help='Create an alias for a role'
    )
    alias_command.add_argument(
        'role', type=str, help='Role name'
    )
    alias_command.add_argument(
        'alias', type=str, help='Alias name'
    )
    return subcommand


def add_info_subcommand(subparsers):
    subcommand = subparsers.add_parser(
        subparsers, 'info',
        help='Print info about data store and config file, etc.',
        formatter_class=ArgumentDefaultsRichHelpFormatter
    )
    subcommand.set_defaults(func=default)
    return subcommand


def add_models_subcommand(subparsers):
    subcommand = subparsers.add_parser(
        'models',
        help='list available models of the backend',
        formatter_class=ArgumentDefaultsRichHelpFormatter
    )
    subcommand.set_defaults(func=models)
    return subcommand


def make_parser():
    parser = argparse.ArgumentParser(
        prog='ae[rgia]', formatter_class=ArgumentDefaultsRichHelpFormatter
    )
    parser.add_argument(
        '--version', action='store_true', help='print the version'
    )
    parser.add_argument(
        '--debug', action='store_true', help='enable debug mode'
    )
    parser.add_argument(
        '--log-level', default=None,
        choices=['debug', 'info', 'warn', 'error', 'critical'],
        help='configure the log-level of the application'
    )

    subparsers = parser.add_subparsers(dest='subcommand')

    add_chat_subcommand(subparsers)
    # add_tui_subcommand(subparsers)
    # add_repl_subcommand(subparsers)
    # add_code_subcommand(subparsers)
    add_image_subcommand(subparsers)
    # add_execute_subcommand(subparsers)
    # add_session_subcommand(subparsers)
    # add_role_subcommand(subparsers)
    # add_info_subcommand(subparsers)
    add_models_subcommand(subparsers)

    return parser



import asyncio
import io
import os
from rich.table import Table
from inspect import cleandoc

import httpx
from openai import AsyncOpenAI, OpenAI
from rich.markdown import Markdown

from enum import IntEnum


class ExitCode(IntEnum):
    Success = 0
    Failure = 1


class TextBuffer:
    """Class to manage a text buffer using StringIO."""

    def __init__(self):
        """
        Initializes a new instance of the TextBuffer class.
        """
        self._buffer = io.StringIO()

    def __enter__(self):
        """
        Enters the runtime context related to this object.

        Returns:
            The TextBuffer instance.
        """
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Exits the runtime context related to this object.

        Args:
            exc_type: The exception type.
            exc_val: The exception value.
            exc_tb: The traceback object.
        """
        self._buffer.close()

    def append(self, text):
        """
        Appends text to the buffer.

        Args:
            text: The text to append to the buffer.
        """
        self._buffer.write(text)

    def __iadd__(self, other):
        """
        Implements the in-place addition operation.

        Args:
            other: The text to add to the buffer.

        Returns:
            The TextBuffer instance.
        """
        self._buffer.write(other)
        return self

    @property
    def content(self):
        """
        Retrieves the content of the buffer.

        Returns:
            The content of the buffer as a string.
        """
        return self._buffer.getvalue()


def default(args, stdout, stderr):
    err_msg = cleandoc(f"""
    This function is not implemented yet!
        
        Details:
            Args: {args}
            
        In case you need more details, consider running the command with the `--debug` flag.
    """)
    raise NotImplementedError(err_msg)


def chat(args, stdout, stderr):
    async def achat(client, model, message):
        stream = await client.chat.completions.create(
            model=model,
            messages=[{"role": "user", "content": message}],
            stream=True,
        )
        with stdout.status("Processing request ...", spinner='aesthetic', spinner_style='cyan') as status:
            with TextBuffer() as buffer:
                async for chunk in stream:
                    text = chunk.choices[0].delta.content or ""
                    buffer.append(text)
                content = buffer.content

        stdout.print(
            Markdown(content)
        )
        return ExitCode.Success

    client = AsyncOpenAI(
        api_key=os.environ.get("OPENAI_API_KEY")
    )
    msg = " ".join(args.text)
    context = "" if not args.context else args.context.read()
    msg = msg + context
    return asyncio.run(achat(client=client, model=args.model, message=msg))


def image(args, stdout, stderr):
    def download(url, file_name):
        resp = httpx.get(url)
        with open(file_name, 'wb') as f:
            f.write(resp.content)

    client = OpenAI(
        api_key=os.environ.get("OPENAI_API_KEY")
    )
    prompt = " ".join(args.text)
    response = client.images.generate(prompt=prompt, model=args.model)
    image_url = response.data[0].url
    download(image_url, args.filename)

    return ExitCode.Success


def models(args, stdout, stderr):
    client = OpenAI(
        api_key=os.environ.get("OPENAI_API_KEY")
    )
    models = client.models.list()

    table = Table(title='Available Models of Backend')
    table.add_column("Id/Name", justify='left', style='green')
    table.add_column("Owned By", justify='left', style='cyan')
    table.add_column("Created", justify='left', style='yellow')

    import datetime
    for m in models:
        created = f"{datetime.datetime.fromtimestamp(m.created)}"
        table.add_row(m.id, m.owned_by, created)

    stdout.print(table)

    return ExitCode.Success



from aergia._logging import logger
from rich.console import Console


class StatusOutput:

    def __init__(self, stderr, logger):
        self._stderr = stderr
        self._logger = logger

    @property
    def stderr(self):
        return self._stderr

    @property
    def logger(self):
        return self._logger

    def print(self, message, *args, **kwargs):
        self._logger.info(message, *args, **kwargs)
        self._stderr.print(message, *args, **kwargs)


stdout = Console()
stderr = StatusOutput(
    stderr=Console(stderr=True),
    logger=logger
)
